<?php

namespace Drupal\acquia_contenthub\Controller;

use Acquia\Hmac\KeyInterface;
use Drupal\acquia_contenthub\Client\ClientFactory;
use Acquia\Hmac\ResponseSigner;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use GuzzleHttp\Psr7\Response;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Bridge\PsrHttpMessage\Factory\DiactorosFactory;
use Drupal\Component\Render\FormattableMarkup;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Component\Serialization\Json;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;

/**
 * Controller for Content Hub Imported Entities.
 */
class ContentHubWebhookController extends ControllerBase {

  /**
   * Logger.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $loggerFactory;

  /**
   * Content Hub Client Factory.
   *
   * @var \Drupal\acquia_contenthub\Client\ClientFactory
   */
  protected $clientFactory;

  /**
   * Drupal Module Handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * @var \Acquia\ContentHubClient\ContentHubClient
   */
  protected $client;

  /**
   * Service Webhooks
   *
   * @var array
   */
  protected $webhooks;

  /**
   * Symfony Request
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected $request;

  /**
   * WebhooksSettingsForm constructor.
   *
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger factory.
   * @param \Symfony\Component\HttpFoundation\Request; $current_request
   *   The client manager.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The Drupal Module Handler.
   */
  public function __construct(LoggerChannelFactoryInterface $logger_factory, Request $current_request, ClientFactory $client_factory, ModuleHandlerInterface $module_handler) {
    $this->loggerFactory = $logger_factory;
    $this->clientFactory = $client_factory;
    $this->client = $this->clientFactory->getClient();
    $this->webhooks = $this->client->getWebHooks();

    // If you're using a host re-writer, you need to find the original host.
    if ($host = $current_request->headers->get('x-original-host')) {
      $current_request->headers->set('host', $host);
    }

    $this->request = $current_request;

    $this->moduleHandler = $module_handler;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    /** @var \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory */
    $logger_factory = $container->get('logger.factory');
    /** @var \Symfony\Component\HttpFoundation\Request $current_request */
    $current_request = $container->get('request_stack')->getCurrentRequest();
    /** @var \Drupal\acquia_contenthub\Client\ClientFactory $client_factory */
    $client_factory = $container->get('acquia_contenthub.client.factory');
    /** @var \Drupal\Core\Extension\ModuleHandlerInterface $module_handler */
    $module_handler = $container->get('module_handler');
    return new static(
      $logger_factory,
      $current_request,
      $client_factory,
      $module_handler
    );
  }

  /**
   * Process a webhook.
   *
   * @return \GuzzleHttp\Psr7\Response
   *   The response object.
   */
  public function receiveWebhook(Request $request) {
    // Obtain the headers.
    $webhook = $request->getContent();
    $response = NULL;

    if ($key = $this->validateWebhookSignature($webhook)) {
      // Notify about the arrival of the webhook request.
      $args = [
        '@whook' => print_r($webhook, TRUE),
      ];
      $message = new FormattableMarkup('Webhook landing: @whook', $args);
      $this->loggerFactory->get('acquia_contenthub')->debug($message);

      if ($webhook = Json::decode($webhook)) {
        // Verification process successful!
        // Now we can process the webhook.
        if (isset($webhook['status'])) {
          switch ($webhook['status']) {
            case 'successful':
              $response = $this->processWebhook($webhook);
              break;

            case 'pending':
              $response = $this->registerWebhook($webhook, $key);
              break;

            default:
              // If any other webhook we are not processing then just display
              // the response.
              $response = new Response();
              break;

          }
        }
      }

    }
    else {
      $ip_address = $this->request->getClientIp();
      $message = new FormattableMarkup('Webhook [from IP = @IP] rejected (Signatures do not match): @whook', [
        '@IP' => $ip_address,
        '@whook' => print_r($webhook, TRUE),
      ]);
      $this->loggerFactory->get('acquia_contenthub')->debug($message);
      $response = new Response();
    }

    return $response;
  }

  /**
   * Validates a webhook signature.
   *
   * @param string $webhook
   *   The Webhook Body.
   *
   * @return bool | \Acquia\Hmac\KeyInterface
   *   TRUE if signature verification passes, FALSE otherwise.
   */
  public function validateWebhookSignature($webhook) {
    $headers = array_map('current', $this->request->headers->all());

    // Quick validation to make sure we are not replaying a request
    // from the past.
    $request_date = isset($headers['date']) ? $headers['date'] : "1970";
    $request_timestamp = strtotime($request_date);
    $timestamp = time();
    // Due to networking delays and mismatched clocks, we are making the request
    // accepting window 60s.
    if (abs($request_timestamp - $timestamp) > 900) {
      $message = new FormattableMarkup('The Webhook request seems that was issued in the past [Request timestamp = @t1, server timestamp = @t2]: rejected: @whook', [
        '@t1' => $request_timestamp,
        '@t2' => $timestamp,
        '@whook' => print_r($webhook, TRUE),
      ]);
      $this->loggerFactory->get('acquia_contenthub')->debug($message);
      return FALSE;
    }

    return $this->clientFactory->authenticate($this->request);
  }

  /**
   * Enables other modules to process the webhook.
   *
   * @param array $webhook
   *   The webhook sent by the Content Hub.
   *
   * @return \Psr\Http\Message\ResponseInterface
   *   The Response Object.
   */
  public function processWebhook(array $webhook) {
    // Process Reindex Webhook.
    if ($webhook['crud'] === 'reindex') {
      //$this->processReindexWebhook($webhook);
      return new Response();
    }
    $assets = isset($webhook['assets']) ? $webhook['assets'] : FALSE;
    if (count($assets) > 0) {
      $this->moduleHandler->alter('acquia_contenthub_process_webhook', $webhook);
    }
    else {
      $message = new FormattableMarkup('Error processing Webhook (It contains no assets): @whook', [
        '@whook' => print_r($webhook, TRUE),
      ]);
      $this->loggerFactory->get('acquia_contenthub')->debug($message);
    }
    return new Response();
  }

  /**
   * Processing the registration of a webhook.
   *
   * @param array $webhook
   *   The webhook coming from Plexus.
   *
   * @param \Acquia\Hmac\KeyInterface $key
   *
   * @return \Psr\Http\Message\ResponseInterface
   *   The Response.
   */
  public function registerWebhook(array $webhook, KeyInterface $key) {
    $uuid = isset($webhook['uuid']) ? $webhook['uuid'] : FALSE;
    $origin = $this->client->getSettings()->getUuid();
    $api_key = $this->client->getSettings()->getApiKey();

    if ($uuid && $webhook['initiator'] == $origin && $webhook['publickey'] == $api_key) {
      $response = new Response();

      $psr7Factory = new DiactorosFactory();
      $psr7_request = $psr7Factory->createRequest($this->request);

      $signer = new ResponseSigner($key, $psr7_request);
      $signedResponse = $signer->signResponse($response);

      return $signedResponse;
    }
    else {
      $ip_address = $this->request->getClientIp();
      $message = new FormattableMarkup('Webhook [from IP = @IP] rejected (initiator and/or publickey do not match local settings): @whook', [
        '@IP' => $ip_address,
        '@whook' => print_r($webhook, TRUE),
      ]);
      $this->loggerFactory->get('acquia_contenthub')->debug($message);
      return new Response();
    }
  }

  /**
   * Process a Reindex Webhook.
   * Todo: this is currently unused from the 1.x port.
   *
   * @param array $webhook
   *   The webhook array.
   */
  private function processReindexWebhook(array $webhook) {
    // Update the Reindex State Variable.
    /** @var \Drupal\acquia_contenthub\Controller\ContentHubReindex $reindex */
    $reindex = \Drupal::service('acquia_contenthub.acquia_contenthub_reindex');
    if ($reindex->isReindexSent()) {
      $reindex->setReindexStateFinished();
    }
  }

}
