<?php

namespace Drupal\acquia_contenthub;

use Acquia\ContentHubClient\CDFDocument;
use Acquia\ContentHubClient\CDFObject;
use Drupal\acquia_contenthub\Event\CDFAttributesEvent;
use Drupal\acquia_contenthub\Event\CreateCDFEntityEvent;
use Drupal\acquia_contenthub\Event\EntityDataTamperEvent;
use Drupal\acquia_contenthub\Event\EntityImportEvent;
use Drupal\acquia_contenthub\Event\ParseCDFEntityEvent;
use Drupal\acquia_contenthub\Event\PruneCDFEntitiesEvent;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Extension\ModuleInstallerInterface;
use Drupal\depcalc\DependencyCalculator;
use Drupal\depcalc\DependentEntityWrapperInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;

/**
 * Serialize an entity to a CDF format.
 *
 * This class will convert an array of entities into a CDF compatible array of
 * data.
 */
class EntityCDFSerializer {

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $dispatcher;

  /**
   * The configuration factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The dependency calculator.
   *
   * @var \Drupal\depcalc\DependencyCalculator
   */
  protected $calculator;

  /**
   * The module handler.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  /**
   * The module installer.
   *
   * @var \Drupal\Core\Extension\ModuleInstallerInterface
   */
  protected $moduleInstaller;

  /**
   * EntityCDFSerializer constructor.
   *
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $dispatcher
   *   The event dispatcher.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The configuration factory.
   * @param \Drupal\depcalc\DependencyCalculator $calculator
   *   The dependency calculator.
   * @param \Drupal\Core\Extension\ModuleInstallerInterface $module_installer
   *   The module installer.
   */
  public function __construct(EventDispatcherInterface $dispatcher, ConfigFactoryInterface $config_factory, DependencyCalculator $calculator, ModuleInstallerInterface $module_installer) {
    $this->dispatcher = $dispatcher;
    $this->configFactory = $config_factory;
    $this->calculator = $calculator;
    $this->moduleInstaller = $module_installer;
  }

  /**
   * Serialize an array of entities into CDF format.
   *
   * @param \Drupal\depcalc\DependentEntityWrapperInterface[] $dependencies
   *   The entity dependency wrappers.
   *
   * @return CDFObject[]
   */
  public function serializeEntities(DependentEntityWrapperInterface ...$dependencies) {
    $output = [];
    foreach ($dependencies as $wrapper) {
      $entity = $wrapper->getEntity();
      $wrapper_dependencies = [];
      if ($entity_dependencies = $wrapper->getDependencies()) {
        $wrapper_dependencies['entity'] = $entity_dependencies;
      }
      if ($module_dependencies = $wrapper->getModuleDependencies()) {
        // Prevent unnecessary string keys.
        $wrapper_dependencies['module'] = array_values($module_dependencies);
      }
      $event = new CreateCDFEntityEvent($entity, $wrapper_dependencies);
      $this->dispatcher->dispatch(AcquiaContentHubEvents::CREATE_CDF_OBJECT, $event);
      foreach ($event->getCDFs() as $cdf) {
        $attributesEvent = new CDFAttributesEvent($cdf, $entity);
        $this->dispatcher->dispatch(AcquiaContentHubEvents::POPULATE_CDF_ATTRIBUTES, $attributesEvent);
        $output[] = $cdf;
      }
    }
    return $output;
  }

  /**
   * Unserializes a cdf into a list of Drupal entities.
   *
   * @todo add more docs about the expected CDF format.
   *
   * @param CDFDocument $cdf
   *   The CDF Document.
   *
   * @param bool $first
   *   Whether or not this is the first invocation.
   *
   * @return CDFDocument
   *   The CDFDocument with all processed application objects.
   * @throws \Exception
   */
  public function unserializeEntities(CDFDocument $cdf, $first = TRUE) {
    if (!$cdf->hasEntities()) {
      throw new \Exception("Missing CDF Entities entry. Not a valid CDF.");
    }
    if ($first) {
      $event = new PruneCDFEntitiesEvent($cdf);
      $this->dispatcher->dispatch(AcquiaContentHubEvents::PRUNE_CDF, $event);
      $cdf = $event->getCDF();
    }
    $this->handleModules($cdf);
    // Allows entity data to be manipulated before unserialization.
    if ($first) {
      $event = new EntityDataTamperEvent($cdf);
      $this->dispatcher->dispatch(AcquiaContentHubEvents::ENTITY_DATA_TAMPER, $event);
      $cdf = $event->getCDF();
    }
    $unordered_entities = $cdf->getEntities();
    // Organize the entities into a dependency chain.
    foreach ($unordered_entities as $uuid => $entity_data) {
      if (!$entity_data->hasApplicationObject() && $this->entityIsProcessable($entity_data, $unordered_entities)) {
        $event = new ParseCDFEntityEvent($entity_data);
        $this->dispatcher->dispatch(AcquiaContentHubEvents::PARSE_CDF, $event);
        $entity = $event->getEntity();
        if ($entity) {
          $entity->save();
          $unordered_entities[$uuid]->setApplicationObject($entity);
          if ($entity->isNew()) {
            $event_name = AcquiaContentHubEvents::ENTITY_IMPORT_NEW;
            $event = new EntityImportEvent($entity, $entity_data);
          }
          else {
            $event_name = AcquiaContentHubEvents::ENTITY_IMPORT_UPDATE;
            $event = new EntityImportEvent($entity, $entity_data);
          }
          $this->dispatcher->dispatch($event_name, $event);
        }
        // @todo provide an api to handle entity data that doesn't resolve into an entity.
        else {
          unset($unordered_entities[$uuid]);
        }
      }
      else {
        unset($unordered_entities[$uuid]);
        $unordered_entities = NestedArray::mergeDeep($unordered_entities, [$uuid => $entity_data]);
      }
    }
    $recurse = FALSE;
    foreach ($unordered_entities as $uuid => $entity) {
      if (!$entity->hasApplicationObject()) {
        $recurse = TRUE;
        break;
      }
    }
    if ($recurse) {
      $cdf->setCDFEntities(...array_values($unordered_entities));
      $cdf = $this->unserializeEntities($cdf, FALSE);
    }
    return $cdf;
  }

  /**
   * Checks dependencies of a cdf entry to determine if it can be processed.
   *
   * CDF entries are turned into Drupal entities. This can only be done when
   * all the dependencies of an entry have been created. This method checks
   * dependencies to ensure they've been properly converted into Drupal
   * entities before proceeding with processing an entry.
   *
   * @param CDFObject $entity_data
   *   The entity data.
   * @param array $unordered_entities
   *   The list of unordered entities processed to this point.
   *
   * @return bool
   */
  protected function entityIsProcessable(CDFObject $entity_data, array $unordered_entities) {
    if (empty($entity_data->getMetadata())) {
      return TRUE;
    }
    foreach ($entity_data->getDependencies() as $key => $dependency_uuid) {
      if (is_numeric($key) && !empty($unordered_entities[$dependency_uuid]) && !($unordered_entities[$dependency_uuid]->hasApplicationObject())) {
        return FALSE;
      }
    }
    return TRUE;
  }

  /**
   * Ensures all required modules of a set of entities are enabled.
   *
   * If modules are missing from the code base, this method will throw an
   * exception before any importing of content can occur which should prevent
   * entities from being in half-operational states.
   *
   * @param \Acquia\ContentHubClient\CDFDocument $cdf
   *   The CDF Document.
   *
   * @throws \Exception The exception thrown if a module is missing from the code base.
   */
  protected function handleModules(CDFDocument $cdf) {
    $dependencies = [];
    $unordered_entities = $cdf->getEntities();
    foreach ($unordered_entities as &$entity) {
      // Don't process entities, their dependencies are working.
      if ($entity->hasApplicationObject()) {
        continue;
      }
      // Don't process non-entities we've previously processed.
      if ($entity->hasProcessedDependencies()) {
        continue;
      }
      // No need to process entities that don't have module dependencies.
      if (!$entity->getModuleDependencies()) {
        continue;
      }
      $dependencies = NestedArray::mergeDeep($dependencies, $entity->getModuleDependencies());
      $entity->markProcessedDependencies();
    }
    $install = FALSE;
    foreach ($dependencies as $module) {
      // @todo consider a configuration that prevents new module installation.
      $install = TRUE;
      // Module isn't installed.
      if (!$this->getModuleHandler()->moduleExists($module)) {
        // Module doesn't exist in the code base, so we can't install.
        if (!drupal_get_filename('module', $module)) {
          // Throw custom exception.
          throw new \Exception(sprintf("The %s module code base is not present.", $module));
        }
      }
      else {
        // Keys and values are the same so this works on this array.
        // Remove dependencies that are already installed.
        unset($dependencies[$module]);
      }
    }
    if ($install) {
      $this->moduleInstaller->install(array_values($dependencies));
      // @todo determine if this cache invalidation is necessary.
      \Drupal::cache()->invalidateAll();
      // Using \Drupal::entityTypeManager() do to caching of the instance in
      // some services. Looks like a core bug.
      \Drupal::entityTypeManager()->clearCachedDefinitions();
    }
  }
  
  protected function getModuleHandler() {
    return \Drupal::moduleHandler();
  }

}
